//
//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// This source code is an unpublished proprietary document, and is available only
// under a written license.
// 

#include <sys/types.h>

#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
// #include <math.h>
#include <ctype.h>

#if defined(GDEBUG)
#define GC_DEBUG
#endif
#include <malloc.h>
#include <signal.h>

#include <xmmintrin.h>
#include <pmmintrin.h>
#include <immintrin.h>

#include "itqA.h"
#include "itq-typedefs.h"
#include "agcfilters.h"


constexpr int MAXTXTBSIZE=256;

constexpr int NUMTHREADS=3;
int numthreads = 0;

//
// Argument processing and storage (thread common) RO on all threads except main and UI
//

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Weffc++"
static struct arg_context acINIT;
struct arg_context ac;
#pragma GCC diagnostic pop


//
// Command line arguements allow for cloud IP to be a DNS name.   However, due to internals of the
// Socketserver interface, the SS IP must be an IP addr.
//
// Other code examples didn't seem to use getopts -- getopts does too many good/nice things to ignore.
//
enum optcode {
	INFCODE, OUTFCODE, INFOCODE, CMDCODE, FLOATOUTCODE,
	THRESHCODE, GAINCODE, OUTGAINCODE, QUIETCODE,
	BOGUSCODE=999999
};

struct optcodenamemap
{
	enum optcode opc;
	const char *name;
};

static optcodenamemap UNUSED optcodenamemap[] =
	{
		{ INFCODE, "INFCODE" }, { OUTFCODE, "OUTFCODE" },
		{ INFOCODE, "INFOCODE" }, { CMDCODE, "CMDCODE" },
		{ GAINCODE, "GAINCODE" }, { THRESHCODE, "THRESHCODE" },
		{ OUTGAINCODE, "OUTGAINCODE" }, { QUIETCODE, "QUIETCODE" },
		{ FLOATOUTCODE, "FLOATOUTCODE"}, { BOGUSCODE, nullptr}
	};

static const char dfltchar='.';

static int pacode;
extern int pdcmd(const char *cmd);

static bool gainset = false;
static float gainval = 0.0f;
static bool thresholdset = false;
static float thresholdval = 0.0f;
static char *previouscmd = nullptr;

static int NOINLINE
process_args(int argc, char *const argv[])
{
	int c, opt_index, __attribute__((unused)) optlen;
	int error = 0;
	static struct option long_options[] = {
		{"outf",	required_argument, &pacode, (int) OUTFCODE},
		{"info",	optional_argument, &pacode, (int) INFOCODE},
		{"inf",	required_argument, &pacode, (int) INFCODE},
		{"gain", required_argument, &pacode, (int) GAINCODE},
		{"threshold", required_argument, &pacode, (int) THRESHCODE},
		{"cmd",	required_argument, &pacode, (int) CMDCODE},
		{"outgain",	required_argument, &pacode, (int) OUTGAINCODE},
		{"floatout",	no_argument, &pacode, (int) FLOATOUTCODE},
		{"quiet",	no_argument, &pacode, (int) QUIETCODE},
		{nullptr,       no_argument      , &pacode, (int) BOGUSCODE}
	};

	//
	// reset scan
	//
//	optind = 1;
	while (1)
	{
		opt_index = -1;
		c = ::getopt_long_only(argc, argv, "", long_options, &opt_index);
#if defined(CMDDEBUG)
		::fprintf(stderr,"getopt returns %d\n", c);
#endif
		if (c == -1)
		{
			break;
		}
		c = pacode;
		const static int optargtempsize=4096;
		enum optcode rtcode = (enum optcode) pacode;
		char optargtemp[optargtempsize];
		memset(optargtemp, 0,sizeof optargtemp);
#if defined(CMDDEBUG)
		::fprintf(stderr,"argument: optarg: %s, opt_index: %d, code: %d, name: \"%s\"\n", optarg, opt_index, rtcode, optlookup(rtcode));
#endif

		int optarglen = 0;
		if (long_options[opt_index].has_arg)
		{
#if defined(CMDDEBUG)
		::fprintf(stderr,"has arg\n");
#endif
			if (optarg)
				optarglen = ::strlen(optarg);
			if (optarglen)
			{
				if (optarglen > (optargtempsize - 1))
				{
					::fprintf(stderr,"argument \"%s\" is too big -- aborting\n", optarg);
					::exit(1);
				}

				// no worry about null terminate - -already handled with prezero above.
				if (optarg[0] == '\"' || optarg[0] == '\'')
				{
					::memcpy(optargtemp, optarg+1, (size_t) optarglen - 2);
				}
				else
				{
					::memcpy(optargtemp, optarg, (size_t) optarglen);
				}
#if defined(CMDDEBUG)
				::fprintf(stderr,"Optarg == \"%s\"\n", optargtemp);
#endif
			}
		}

#if defined(CMDDEBUG)
		::fprintf(stderr,"rtcode: %d\n", rtcode);
#endif

		switch (rtcode)
		{

		case CMDCODE:
			{
				if (optargtemp[0] == '\0')
					break;
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"new --cmd=%s will be ignored, previous --cmd=\"%s\" will be used\n",
						optargtemp, previouscmd);
					break;
				}
				if (gainset || thresholdset)
				{
					fprintf(stderr,"gain or threshold already set, new --cmd=\"%s\" values will override\n", optargtemp);
					gainset = false;
					thresholdset = false;
				}
				if (ac.info > 3) fprintf(stderr,"cmd=\"%s\"\n", optargtemp);

				if (previouscmd != nullptr)
				{
					::free(previouscmd);
				}

				previouscmd = ::strdup(optargtemp);
				
				pdcmd(optargtemp);
			}
			break;

		case GAINCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --gain\n");
					break;
				}
				float gaintmp = atof(optargtemp);
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--gain=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
					
				if (gainset)
				{
					fprintf(stderr,"--gain switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						gainval, gaintmp);
				}
				gainset = true;
				gainval = gaintmp;
			}
			break;

		case OUTGAINCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --outgain\n");
					break;
				}
				float gaintmp = atof(optargtemp);
				if (gaintmp > 20.0f || gaintmp < -20.0f)
				{
					fprintf(stderr,"Warning: eccentric output gain given (--outgain=\"%s\") -- -20.0dB to +20.0dB more reasonable\n", optargtemp);
				}
				ac.outgain = dB2lin(gaintmp);
			}
			break;

		case THRESHCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the threshold\n");
					break;
				}
				float threshtmp = atof(optargtemp);
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--threshold=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
					
				if (thresholdset)
				{
					fprintf(stderr,"--threshold switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						thresholdval, threshtmp);
				}
				thresholdset = true;
				thresholdval = threshtmp;
			}
			break;
			
		//
		// input file
		//
		case INFCODE:
			
			if (optargtemp[0] == '\0')
				break;
#if defined(CMDDEBUG)
		::fprintf(stderr,"optargtemp: %s\n", optargtemp);
#endif
			if (ac.infile != nullptr)
				::free(ac.infile);
			ac.infile = ::strdup(optargtemp);
#if defined(CMDDEBUG)
		::fprintf(stderr,"new infile: %s\n", ac.infile);
#endif
			break;

		case QUIETCODE:
			ac.quiet = true;
			break;

		//
		// cl_port
		//
		case OUTFCODE:
			if (optargtemp[0] == '\0')
				break;
#if defined(CMDDEBUG)
		::fprintf(stderr,"optargtemp: %s\n", optargtemp);
#endif
			if (ac.outfile != nullptr)
				::free(ac.outfile);
			ac.outfile = ::strdup(optargtemp);
#if defined(CMDDEBUG)
		::fprintf(stderr,"new outfile: %s\n", ac.outfile);
#endif
			break;

		//
		// loud debug
		//
		case INFOCODE:
			if (optargtemp[0] == '\0')
			{
				ac.info = 1;
				break;
			}
			ac.info = atoi(optargtemp);
fprintf(stderr,"setting info to be: %d\n", ac.info);
			break;

		case BOGUSCODE:
			break;

		case FLOATOUTCODE:
			ac.rafloatout = 1;
			if (ac.info > 0)
				::fprintf(stderr,"32 bit floating point output: enabled\n");
			break;
			
		}
	}

	if (previouscmd == nullptr)
	{
		if (!thresholdset)
		{
			thresholdval = -3.0f;
			if (!ac.quiet) fprintf(stderr,"Threshold not set, assuming (incorrectly) that threshold = %f\n", thresholdval);
		}
		if (!gainset)
		{
			gainval = 0.0f;
			if (!ac.quiet || (ac.info > 4)) fprintf(stderr,"Gain not set, assuming (probably correctly) that the gain should be = %f\n", gainval);
		}
		char previouscmd[2048];
		int nc = snprintf(previouscmd, sizeof previouscmd, "(l,%f,%f)", thresholdval, gainval);
		if (nc != -1)
		{
			if (ac.info)
			{
				if (!ac.quiet) fprintf(stderr,"using threshold = %f, gain = %f\n", thresholdval, gainval);
			}
			pdcmd(previouscmd);
		}
	}

	return error;
}

extern int setupinfile(void);

static char inbuf[16384];
static char outbuf[16384];

static void displayconfig(FILE *);

static void
wakesig(int __attribute__((unused)) sig)
{
	return;
}

extern void runexp(void);

pid_t mainpid = 0;
static void acinit();
int
main(int argc, char * const argv[])
{

#if !defined(WINDOWS64)
	fcntl(0, F_SETPIPE_SZ, 262144);

	fcntl(1, F_SETPIPE_SZ, 262144);
#endif

	mainpid = getpid();
	acinit();

	process_args(argc, argv);

	if (!ac.quiet || (ac.info > 4))
	{
		fprintf(stderr,"Audio expander V1.0 -- Author John S. Dyson -- Copyright 2017\nLicense terms are in accompanying documentation\n");
		if (ac.info) fprintf(stderr,"Required CPU type: %s\n", VNAME);
	}


	if (ac.info == 0)
	{
//		::fprintf(stderr,"Some help:  NEEDS TO BE COMPLETED AGAIN\n");
	}

	if ((ac.infile == nullptr) || !::strncmp(ac.infile, "-", 2))
	{
		ac.infp = fdopen(dup(0), "rb");
		if (ac.infile == nullptr)
			ac.infile = strdup("Using stdin");
	}
	else
	{
		ac.infp = fopen(ac.infile, "rb");
	}

	if (ac.infp == nullptr)
	{
#if !defined(NOSTRERROR)
		char sbuf[MAXTMPSTR + 1];
		const char *rtstr = lstrerror_r(errno, sbuf, sizeof sbuf);
		::fprintf(stderr,"open error \"%s\", code: %d, for input file: \"%s\"\n", rtstr, errno, ac.infile);
#else
		const char *rtstr = strerror(errno);
		::fprintf(stderr,"open error \"%s\", code: %d, for input file: \"%s\"\n", rtstr, errno, ac.infile);
#endif
		::exit(1);
	}

#if defined(CMDDEBUG)
	::fprintf(stderr, "trying to open output -- \"%s\"\n", ac.outfile);
#endif

	if (ac.outfile == nullptr)
	{
		ac.outfp = fdopen(dup(1), "wb");
#if defined(CMDDEBUG)
	::fprintf(stderr, "duped output fd: 1\n");
#endif
	}
	else
	{
		ac.outfp = fopen(ac.outfile, "w+b");
	}

#if defined(CMDDEBUG)
	::fprintf(stderr, "opened output -- \"%s\", fd: %d\n", ac.outfile, fileno(ac.outfp));
#endif
	if (ac.outfp == nullptr)
	{
#if !defined(NOSTRERROR)
		char sbuf[MAXTMPSTR + 1];
		const char *rtstr = lstrerror_r(errno, sbuf, sizeof sbuf);
		::fprintf(stderr,"open error \"%s\", code: %d, for create/output file: \"%s\"\n", rtstr, errno, ac.outfile);
#else
		const char *rtstr = strerror(errno);
		::fprintf(stderr,"open error \"%s\", code: %d, for create/output file: \"%s\"\n", rtstr, errno, ac.outfile);
#endif
		::exit(1);
	}

	setvbuf(ac.infp, inbuf, _IOFBF, sizeof inbuf);
	setvbuf(ac.outfp, outbuf, _IOFBF, sizeof outbuf);

#if defined(CMDDEBUG)
	::fprintf(stderr, "file %s open with fd %d, file %s open with fd %d\n",
		ac.infile, fileno(ac.infp), ac.outfile, fileno(ac.outfp));
#endif

	if (setupinfile())
	{
		fprintf(stderr,"Failure to start -- input file is incompatible\n");
		exit(1);
	}

	setsignal();
	forcettyupdate();
	accuminfo::initaccumulatelevel();
	runexp();

	deletefilters();
	::exit(0);
}

static void NOINLINE
acinit(void)
{
	setfpmode();
	acINIT.infile = nullptr;
	acINIT.outfile = nullptr;
	acINIT.infp = nullptr;
	acINIT.outfp = nullptr;
	acINIT.rafloatout = -1;
	acINIT.info = 0;
	acINIT.outgain = 1.0f;
	acINIT.quiet = false;

	ac = acINIT;
}


//
// vi: set ts=4:
//
