
//
//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// This source code is an unpublished proprietary document, and is available only
// under a written license.
// 

#include <sys/types.h>


#if !defined(__VECMATH_H__)

#define __VECMATH_H__

#if 1
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Weffc++"
#include "vecmathlib.h"
#pragma GCC diagnostic pop
#endif
#include "varray.h"

#if !defined(__AVX__)
#define OPT4SIMPLE
#endif

#if !defined(UNUSED)
#define UNUSED __attribute__((unused))
#endif

typedef float v8sf __attribute__((vector_size(32)));
typedef float v4sf __attribute__((vector_size(16)));

#define VECSIZE VECMATHLIB_MAX_FLOAT_VECSIZE

#if defined(LIMITVECSIZE)

#if LIMITVECSIZE > VECMATHLIB_MAX_FLOAT_VECSIZE

#undef LIMITVECSIZE
#define LIMITVECSIZE VECMATHLIB_MAX_FLOAT_VECSIZE

#endif

#undef VECSIZE
#define VECSIZE LIMITVECSIZE

typedef vecmathlib::realvec<double, VECSIZE/2> float64v;
typedef vecmathlib::realvec<float, VECSIZE> float32v;
typedef vecmathlib::intvec<float, VECSIZE> int32v;
typedef vecmathlib::boolvec<float, VECSIZE> bool32v;


#else

typedef vecmathlib::realvec<double, VECSIZE/2> float64v;
typedef vecmathlib::realvec<float, VECSIZE> float32v;
typedef vecmathlib::intvec<float, VECSIZE> int32v;
typedef vecmathlib::boolvec<float, VECSIZE> bool32v;

#endif

typedef vecmathlib::realvec<double, 2> float64d2;
typedef vecmathlib::realvec<double, 4> float64d4;
typedef vecmathlib::realvec<float, 4> float32f4;
typedef vecmathlib::realvec<float, 8> float32f8;

#if VECSIZE >= 8
#define USEV8SF
#endif


typedef vecmathlib::realvec<float, 1> float1;
typedef vecmathlib::realvec<double, 1> double1;

static constexpr double VECMATHSMALLVAL = 1.0e-7;
static constexpr float VECMATHSMALLVALF = float(VECMATHSMALLVAL);

static constexpr double LN10 = 2.30258509299404568402;
static constexpr float LN10F = float(LN10);

using namespace std;
using namespace vecmathlib;

static inline float __attribute__((pure))
vpowf(float in, float p)
{
	//typedef vecmathlib::realvec<float,4> vtype;
	typedef float32_vec vtype;
	typedef mathfuncs<vtype> mf;
	return mf::vml_pow(vtype(in), vtype(p))[0];
}

template <typename DT>
static DT
log10(DT src)
{
	return log(src)/DT(LN10);
}

template <typename DT>
static DT
exp10(DT src)
{
	return exp(src * DT(LN10));
}

#if defined(__AVX__)

static varrayf8 UNUSED 
log(varrayf8 inval)
{
	typedef vecmathlib::realvec<float,8> vtype;
	typedef mathfuncs<vtype> mf;

	vtype rtval = mf::vml_log(vtype(inval));
	return varrayf8(rtval);
}

static varrayd4 UNUSED
log(varrayd4 inval)
{
	using vtype = vecmathlib::realvec<double,4>;
	using mf = mathfuncs<vtype>;

	vtype rtval = mf::vml_log(vtype(inval));

	return varrayd4(rtval);
}

static varrayd4 UNUSED
exp(varrayd4 inval)
{
	using vtype = vecmathlib::realvec<double,4>;
	using mf = mathfuncs<vtype>;

	vtype rtval = mf::vml_exp(vtype(inval));

	return varrayd4(rtval);
}

static varrayd4 UNUSED 
pow(varrayd4 inval, double powval)
{
	typedef vecmathlib::realvec<double,4> vtype4;
	typedef mathfuncs<vtype4> mf;

	varrayd4 pv = (varrayd4) {powval, powval, powval, powval};
	vtype4 rtval = mf::vml_pow(vtype4(inval), vtype4(pv));
	return varrayd4(rtval);
}

static varrayd4 UNUSED
pow(varrayd4 inval, varrayd4 powval)
{
	typedef vecmathlib::realvec<double,4> vtype4;
	typedef mathfuncs<vtype4> mf;

	vtype4 rtval = mf::vml_pow(vtype4(inval), vtype4(powval));

	return varrayd4(rtval);
}

static varrayf8 UNUSED 
pow(varrayf8 inval, float powval)
{
	using vtype = vecmathlib::realvec<float,8>;
	using mf = mathfuncs<vtype>;

	varrayf8 pv = (varrayf8) {powval, powval, powval, powval, powval, powval, powval, powval};

	vtype rtval = mf::vml_pow(vtype(inval), vtype(pv));
	return varrayf8(rtval);
}

static varrayf8 UNUSED
pow(varrayf8 inval, varrayf8 powval)
{
	using vtype = vecmathlib::realvec<float,8>;
	using mf = mathfuncs<vtype>;

	vtype rtval = mf::vml_pow(vtype(inval), vtype(powval));

	return varrayf8(rtval);
}

static varrayd8 UNUSED
pow(varrayd8 inval, double powval)
{
	using vtype = vecmathlib::realvec<double,4>;
	using mf = mathfuncs<vtype>;

	v4df pv = (v4df) {powval, powval, powval, powval};
	v4df iv0{};
	v4df iv1{};
	for (int i = 0; i < 4; i++)
		iv0[i] = inval[i];
	for (int i = 0; i < 4; i++)
		iv1[i] = inval[i+4];

	vtype rt0 = mf::vml_pow(vtype(iv0), vtype(pv));
	vtype rt1 = mf::vml_pow(vtype(iv1), vtype(pv));

	varrayd8 rtval{};

	for (int i = 0; i < 4; i++)
		rtval[i] = rt0[i];
	for (int i = 0; i < 4; i++)
		rtval[i+4] = rt1[i];

	return rtval;
}

static varrayf8 UNUSED 
lin2dB(varrayf8 src)
{
	varrayf8 smallvalv8(VECMATHSMALLVALF);
	varrayf8 twentyv8(20.0f);

	varrayf8 tmp = Vmax(src, smallvalv8);
	varrayf8 rtval = varrayf8(log10(float32v(tmp)));
	rtval *= twentyv8;

	return rtval;
}

static varrayf8 UNUSED 
dB2lin(varrayf8 src)
{
	varrayf8 invtwentyv8(0.05f);
	varrayf8 tmp = src;
	tmp *= invtwentyv8;
	return varrayf8(exp10(float32v(tmp)));
}

static varrayd4 UNUSED 
lin2dB(varrayd4 src)
{
	varrayd4 smallvalv4(VECMATHSMALLVAL);
	varrayd4 twentyv4(20.0);

	varrayd4 tmp = Vmax(src, smallvalv4);
	varrayd4 rtval = varrayd4(log10(float64d4(tmp)));
	rtval *= twentyv4;

	return rtval;
}

static varrayd4 UNUSED 
dB2lin(varrayd4 src)
{
	varrayd4 invtwentyv4(0.05);
	varrayd4 tmp(src);
	tmp *= invtwentyv4;
	return varrayd4(exp10(float64d4(tmp)));
}

static varrayf4 UNUSED 
lin2dB(varrayf4 src)
{
	return varrayf4(lin2dB(varrayf4(src)));
}

static varrayf4 UNUSED 
dB2lin(varrayf4 src)
{
	return varrayf4(dB2lin(varrayf4(src)));
}

static float UNUSED 
lin2dB(float srcv)
{
	typedef mathfuncs<float32_vec> mf;
	float srcvtmp = (srcv > 0.0001f) ? srcv : 0.0001f;
	return 20.0f * mf::vml_log10(float32_vec(srcvtmp))[0];
}

static float UNUSED 
dB2lin(float srcv)
{
	typedef mathfuncs<float32_vec> mf;
	static constexpr float div20 = 0.05f;
	return mf::vml_exp10(float32_vec(srcv * div20))[0];
}

static double UNUSED 
lin2dB(double srcv)
{
	typedef mathfuncs<float64_vec> mf;
	double srcvtmp = (srcv > 0.0001) ? srcv : 0.0001;
	return 20.0 * mf::vml_log10(float64_vec(srcvtmp))[0];
}

static double UNUSED 
dB2lin(double srcv)
{
	typedef mathfuncs<float64_vec> mf;
	static constexpr double div20 = 0.05;
	return mf::vml_exp10(float64_vec(srcv * div20))[0];
}

static varrayd8 UNUSED 
lin2dB(varrayd8 src)
{
	varrayd8 smallvalv8(VECMATHSMALLVAL);
	varrayd8 twentyv8(20.0f);

	varrayd8 tmp = Vmax(src, smallvalv8);
	varrayd8 rtval;

	rtval.u.vdata2[0] = log10(float64v(tmp.u.vdata2[0]));
	rtval.u.vdata2[1] = log10(float64v(tmp.u.vdata2[1]));
	rtval *= twentyv8;

	return rtval;
}

static varrayd8 UNUSED 
dB2lin(varrayd8 src)
{
	varrayd8 invtwentyv8(0.05);
	varrayd8 tmp = src * invtwentyv8;

	varrayd8 rtval{0.0};

	rtval.u.vdata2[0] = exp10(float64v(tmp.u.vdata2[0]));
	rtval.u.vdata2[1] = exp10(float64v(tmp.u.vdata2[1]));
	
	return rtval;
}

#elif defined(__SSE2__)

static varrayf8 UNUSED 
log(varrayf8 inval)
{
	using vtype = vecmathlib::realvec<float,4>;
	using mf = mathfuncs<vtype>;

	varrayf4 iv0, iv1;
	for (int i = 0; i < 4; i++)
		iv0[i] = inval[i];
	for (int i = 0; i < 4; i++)
		iv1[i] = inval[i+4];

	vtype rt0 = mf::vml_log(vtype(iv0));
	vtype rt1 = mf::vml_log(vtype(iv1));

	varrayf8 rtval;

	for (int i = 0; i < 4; i++)
		rtval[i] = rt0[i];
	for (int i = 0; i < 4; i++)
		rtval[i+4] = rt1[i];

	return rtval;
}

static varrayd4 UNUSED
pow(varrayd4 inval, varrayd4 powval)
{
	using vtype = vecmathlib::realvec<double,2>;
	using mf = mathfuncs<vtype>;

	varrayd2 inval0tmp{inval[0], inval[1]};
	varrayd2 powval0tmp{powval[0], powval[1]};
	varrayd2 rtval0tmp{mf::vml_pow(vtype(inval0tmp), vtype(powval0tmp))};

	varrayd2 inval1tmp{inval[2], inval[3]};
	varrayd2 powval1tmp{powval[2], powval[3]};
	varrayd2 rtval1tmp{mf::vml_pow(vtype(inval1tmp), vtype(powval1tmp))};

	return varrayd4{rtval0tmp[0], rtval0tmp[1], rtval1tmp[0], rtval1tmp[1]};
}

static varrayd4 UNUSED
pow(varrayd4 inval, double powvalA)
{
	varrayd4 powval(powvalA);

	return pow(inval, powval);
}

static varrayd4 UNUSED
log(varrayd4 inval)
{
	using vtype = vecmathlib::realvec<double,2>;
	using mf = mathfuncs<vtype>;

	varrayd2 inval0tmp{inval[0], inval[1]};
	varrayd2 rtval0tmp{mf::vml_log(vtype(inval0tmp))};

	varrayd2 inval1tmp{inval[2], inval[3]};
	varrayd2 rtval1tmp{mf::vml_log(vtype(inval1tmp))};

	return varrayd4{rtval0tmp[0], rtval0tmp[1], rtval1tmp[0], rtval1tmp[1]};
}

static varrayd4 UNUSED
exp(varrayd4 inval)
{
	using vtype = vecmathlib::realvec<double,2>;
	using mf = mathfuncs<vtype>;

	varrayd2 inval0tmp{inval[0], inval[1]};
	varrayd2 rtval0tmp{mf::vml_exp(vtype(inval0tmp))};

	varrayd2 inval1tmp{inval[2], inval[3]};
	varrayd2 rtval1tmp{mf::vml_exp(vtype(inval1tmp))};

	return varrayd4{rtval0tmp[0], rtval0tmp[1], rtval1tmp[0], rtval1tmp[1]};
}

static varrayf8 UNUSED 
pow(varrayf8 inval, float powval)
{
	using vtype = vecmathlib::realvec<float,4>;
	using mf = mathfuncs<vtype>;

	varrayf4 pv = (varrayf4) {powval, powval, powval, powval};
	varrayf4 iv0, iv1;
	for (int i = 0; i < 4; i++)
		iv0[i] = inval[i];
	for (int i = 0; i < 4; i++)
		iv1[i] = inval[i+4];

	vtype rt0 = mf::vml_pow(vtype(iv0), vtype(pv));
	vtype rt1 = mf::vml_pow(vtype(iv1), vtype(pv));

	varrayf8 rtval;

	for (int i = 0; i < 4; i++)
		rtval[i] = rt0[i];
	for (int i = 0; i < 4; i++)
		rtval[i+4] = rt1[i];

	return rtval;
}

static varrayf8 UNUSED
pow(varrayf8 inval, varrayf8 powval)
{
	using vtype = vecmathlib::realvec<float,4>;
	using mf = mathfuncs<vtype>;

	varrayf4 pv0 = (varrayf4) {powval[0], powval[1], powval[2], powval[3]};
	varrayf4 pv1 = (varrayf4) {powval[4], powval[5], powval[6], powval[7]};
	varrayf4 iv0, iv1;
	for (int i = 0; i < 4; i++)
		iv0[i] = inval[i];
	for (int i = 0; i < 4; i++)
		iv1[i] = inval[i+4];

	vtype rt0 = mf::vml_pow(vtype(iv0), vtype(pv0));
	vtype rt1 = mf::vml_pow(vtype(iv1), vtype(pv1));

	varrayf8 rtval;

	for (int i = 0; i < 4; i++)
		rtval[i] = rt0[i];
	for (int i = 0; i < 4; i++)
		rtval[i+4] = rt1[i];

	return rtval;
}

static varrayd8 UNUSED
pow(varrayd8 inval, double powval)
{

	using vtype = vecmathlib::realvec<double,2>;
	using mf = mathfuncs<vtype>;

	varrayd2 pv = (varrayd2) {powval, powval};
	varrayd2 iv0, iv1, iv2, iv3;

	iv0[0] = inval[0];
	iv0[1] = inval[1];
	iv1[0] = inval[2];
	iv1[1] = inval[3];
	iv2[0] = inval[4];
	iv2[1] = inval[5];
	iv3[0] = inval[6];
	iv3[1] = inval[7];

	vtype rt0 = mf::vml_pow(vtype(iv0), vtype(pv));
	vtype rt1 = mf::vml_pow(vtype(iv1), vtype(pv));
	vtype rt2 = mf::vml_pow(vtype(iv2), vtype(pv));
	vtype rt3 = mf::vml_pow(vtype(iv3), vtype(pv));

	varrayd8 rtval;

	rtval[0] = rt0[0];
	rtval[1] = rt0[1];
	rtval[2] = rt1[0];
	rtval[3] = rt1[1];
	rtval[4] = rt2[0];
	rtval[5] = rt2[1];
	rtval[6] = rt3[0];
	rtval[7] = rt3[1];

	return rtval;
}


static varrayf8 UNUSED
lin2dB(varrayf8 src)
{
	using vtype = vecmathlib::realvec<float,4>;
	using mf = mathfuncs<vtype>;

	varrayf4 smallvalv4(VECMATHSMALLVALF);
	varrayf4 twentyv4(20.0f);

	varrayf4 tmp0 { src[0], src[1], src[2], src[3] };
	tmp0 = Vmax( tmp0, smallvalv4);
	varrayf4 rtval0 { log10(vtype(tmp0)) };
	rtval0 *= twentyv4;

	varrayf4 tmp1 { src[4], src[5], src[6], src[7] };
	tmp1 = Vmax( tmp1, smallvalv4);
	varrayf4 rtval1 { log10(vtype(tmp1)) };
	rtval1 *= twentyv4;

	return varrayf8 {
		rtval0[0], rtval0[1], rtval0[2], rtval0[3],
		rtval1[0], rtval1[1], rtval1[2], rtval1[3]
	};
}

static varrayf8 UNUSED
dB2lin(varrayf8 src)
{
	using vtype = vecmathlib::realvec<float,4>;
	using mf = mathfuncs<vtype>;

	varrayf4 invtwentyv4(0.05f);

	varrayf4 tmp0 { src[0], src[1], src[2], src[3] };
	tmp0 *= invtwentyv4;
	varrayf4 rtval0 { exp10(vtype(tmp0)) };

	varrayf4 tmp1 { src[0], src[1], src[2], src[3] };
	tmp1 *= invtwentyv4;
	varrayf4 rtval1 { exp10(vtype(tmp1)) };

	return varrayf8 {
		rtval0[0], rtval0[1], rtval0[2], rtval0[3],
		rtval1[0], rtval1[1], rtval1[2], rtval1[3]
	};
}

static varrayd4 UNUSED
lin2dB(varrayd4 src)
{
	using vtype = vecmathlib::realvec<double,2>;
	using mf = mathfuncs<vtype>;

	varrayd2 smallvalv2(VECMATHSMALLVAL);
	varrayd2 twentyv2(20.0);

	varrayd2 tmp0 { src[0], src[1] };
	tmp0 = Vmax( tmp0, smallvalv2);
	varrayd2 rtval0 { log10(vtype(tmp0)) };
	rtval0 *= twentyv2;

	varrayd2 tmp1 { src[2], src[3] };
	tmp1 = Vmax( tmp1, smallvalv2);
	varrayd2 rtval1 { log10(vtype(tmp1)) };
	rtval1 *= twentyv2;

	return { rtval0[0], rtval0[1], rtval1[0], rtval1[1] };
}

static varrayd4 UNUSED
dB2lin(varrayd4 src)
{
	using vtype = vecmathlib::realvec<double,2>;
	using mf = mathfuncs<vtype>;

	varrayd2 invtwentyv2(0.05);

	varrayd2 tmp0 { src[0], src[1] };
	tmp0 *= invtwentyv2;
	varrayd2 rtval0 { exp10(vtype(tmp0)) };

	varrayd2 tmp1 { src[2], src[3] };
	tmp1 *= invtwentyv2;
	varrayd2 rtval1 { exp10(vtype(tmp1)) };

	return { rtval0[0], rtval0[1], rtval1[0], rtval1[1] };
}

static varrayf4 UNUSED
lin2dB(varrayf4 src)
{
	return varrayf4(lin2dB(varrayd4(src)));
}

static varrayf4 UNUSED
dB2lin(varrayf4 src)
{
	return varrayf4(dB2lin(varrayd4(src)));
}

static float UNUSED
lin2dB(float srcv)
{
	typedef mathfuncs<float1> mf;
	float srcvtmp = (srcv > 0.0001) ? srcv : 0.0001;
	return 20.0f * mf::vml_log10(srcvtmp);
}

static float UNUSED
dB2lin(float srcv)
{
	typedef mathfuncs<float1> mf;
	const static float div20 = 0.05f;
	return mf::vml_exp10(srcv * div20);
}


static varrayd8 UNUSED
lin2dB(varrayd8 src)
{
	varrayd4 tmp0 { src[0], src[1], src[2], src[3] };
	varrayd4 rtval0 = lin2dB(tmp0);

	varrayd4 tmp1 { src[4], src[5], src[6], src[7] };
	varrayd4 rtval1 = lin2dB(tmp1);

	return varrayd8 {
		rtval0[0], rtval0[1], rtval0[2], rtval0[3],
		rtval1[0], rtval1[1], rtval1[2], rtval1[3]
	};
}

static varrayd8 UNUSED
dB2lin(varrayd8 src)
{
	varrayd4 tmp0 { src[0], src[1], src[2], src[3] };
	varrayd4 rtval0 = dB2lin(tmp0);

	varrayd4 tmp1 { src[4], src[5], src[6], src[7] };
	varrayd4 rtval1 = dB2lin(tmp1);

	return varrayd8 {
		rtval0[0], rtval0[1], rtval0[2], rtval0[3],
		rtval1[0], rtval1[1], rtval1[2], rtval1[3]
	};
}

#else

static varrayf8 UNUSED 
log(varrayf8 inval)
{
	varrayf8 rtval;
	for (int i = 0; i < 8; i++)
	{
		rtval[i] = ::log(inval[i]);
	}
	return rtval;
}

static varrayd4 UNUSED
log(varrayd4 inval)
{
	using vtype = vecmathlib::realvec<double,1>;
	using mf = mathfuncs<vtype>;

	varrayd4 rtval;
	for (int i = 0; i < 4; i++)
		rtval[i] = mf::vml_log(vtype(inval[i]));
	return rtval;
}

static varrayd4 UNUSED
exp(varrayd4 inval)
{
	using vtype = vecmathlib::realvec<double,1>;
	using mf = mathfuncs<vtype>;

	varrayd4 rtval;
	for (int i = 0; i < 4; i++)
		rtval[i] = mf::vml_exp(vtype(inval[i]));
	return rtval;
}

static varrayf8 UNUSED 
pow(varrayf8 inval, float powval)
{
	varrayf8 rtval;
	for (int i = 0; i < 8; i++)
	{
		rtval[i] = ::pow(inval[i], powval);
	}
	return rtval;
}

static varrayf8 UNUSED
pow(varrayf8 inval, varrayf8 powval)
{
	varrayf8 rtval;
	for (int i = 0; i < 8; i++)
	{
		rtval[i] = ::pow(inval[i], powval[i]);
	}
	return rtval;
}

static varrayd8 UNUSED
pow(varrayd8 inval, double powval)
{
	varrayd8 rtval;
	for (int i = 0; i < 8; i++)
	{
		rtval[i] = ::pow(inval[i], powval);
	}
	return rtval;
}

LDBCVTYPE varrayf8 LDBCVATTR
lin2dB(varrayf8 src)
{
	varrayf4 small(VECMATHSMALLVALF);
	varrayf4 twenty(20.0f);

	varrayf4 tmp{};

	for (int i = 0; i < 4; i++)
		tmp[i] = src[i];
	tmp = Vmax(tmp,small);
	varrayf4 rtval0{log10(float32v(tmp))};
	rtval0 *= twenty;

	tmp[0] = src[4];
	tmp[1] = src[5];
	tmp[2] = src[6];
	tmp[3] = src[7];

	tmp = Vmax(tmp, small);
	varrayf4 rtval1{log10(float32v(tmp))};
	rtval1 *= twenty;

	varrayf8 rtval;
	for (int i = 0; i < 4; i++)
		rtval[i] = rtval0[i];
	for (int i = 0; i < 4; i++)
		rtval[i+4] = rtval1[i];

	return rtval;
}

LDBCVTYPE varrayd8 LDBCVATTR
lin2dB(varrayd8 src)
{
	varrayd2 tmp{};
	varrayd2 small(VECMATHSMALLVAL);
	varrayd2 twenty(20.0);

	tmp[0] = src[0];
	tmp[1] = src[1];
	tmp = Vmax(tmp,small);
	varrayd2 rtval0{log10(float64v(tmp))};
	rtval0 *= twenty;

	tmp[0] = src[2];
	tmp[1] = src[3];
	tmp = Vmax(tmp,small);
	varrayd2 rtval1{log10(float64v(tmp))};
	rtval1 *= twenty;

	tmp[0] = src[4];
	tmp[1] = src[5];
	tmp = Vmax(tmp,small);
	varrayd2 rtval2{log10(float64v(tmp))};
	rtval2 *= twenty;

	tmp[0] = src[6];
	tmp[1] = src[7];
	tmp = Vmax(tmp,small);
	varrayd2 rtval3{log10(float64v(tmp))};
	rtval3 *= twenty;

	varrayd8 rtval;
	rtval[0] = rtval0[0];
	rtval[1] = rtval0[1];
	rtval[2] = rtval1[0];
	rtval[3] = rtval1[1];
	rtval[4] = rtval2[0];
	rtval[5] = rtval2[1];
	rtval[6] = rtval3[0];
	rtval[7] = rtval3[1];

	return rtval;
}

static varrayf8
dB2lin(varrayf8 src)
{
	varrayf4 tmp{};
	varrayf4 invtwenty(0.05f);

	for (int i = 0; i < 4; i++)
		tmp[i] = src[i];
	tmp *= invtwenty;
	varrayf4 rtval0{exp10(float32v(tmp))};

	for (int i = 0; i < 4; i++)
		tmp[i] = src[i+4];
	tmp *= invtwenty;
	varrayf4 rtval1{exp10(float32v(tmp))};
	
	varrayf8 rtval;
	for (int i = 0; i < 4; i++)
		rtval[i] = rtval0[i];
	for (int i = 0; i < 4; i++)
		rtval[i+4] = rtval1[i];

	return rtval;
}

static varrayd8
dB2lin(varrayd8 src)
{
	varrayd2 tmp{};
	varrayd2 invtwenty(0.05f);

	tmp[0] = src[0];
	tmp[1] = src[1];
	tmp *= invtwenty;
	varrayd2 rtval0{exp10(float64v(tmp))};

	tmp[0] = src[2];
	tmp[1] = src[3];
	tmp *= invtwenty;
	varrayd2 rtval1{exp10(float64v(tmp))};

	tmp[0] = src[4];
	tmp[1] = src[5];
	tmp *= invtwenty;
	varrayd2 rtval2{exp10(float64v(tmp))};

	tmp[0] = src[6];
	tmp[1] = src[7];
	tmp *= invtwenty;
	varrayd2 rtval3{exp10(float64v(tmp))};

	varrayd8 rtval;
	rtval[0] = rtval0[0];
	rtval[1] = rtval0[1];
	rtval[2] = rtval1[0];
	rtval[3] = rtval1[1];
	rtval[4] = rtval2[0];
	rtval[5] = rtval2[1];
	rtval[6] = rtval3[0];
	rtval[7] = rtval3[1];

	return rtval;
}


static varrayd4 UNUSED
lin2dB(varrayd4 src)
{
	varrayd4 smallvalv4(VECMATHSMALLVAL);
	varrayd4 twentyv4(20.0);

	varrayd4 tmp = Vmax(src, smallvalv4);
	varrayd4 rtval = varrayd4(log10(tmp));
	rtval *= twentyv4;

	return rtval;
}

static varrayd4 UNUSED
dB2lin(varrayd4 src)
{
	varrayd4 invtwentyv4(0.05);
	varrayd4 tmp(src);
	tmp *= invtwentyv4;
	return varrayd4(exp10(tmp));
}

static varrayf4 UNUSED
lin2dB(varrayf4 src)
{
	return varrayf4(lin2dB(varrayd4(src)));
}

static varrayf4 UNUSED
dB2lin(varrayf4 src)
{
	return varrayf4(dB2lin(varrayd4(src)));
}

static float UNUSED
lin2dB(float srcv)
{
	typedef mathfuncs<float1> mf;
	return 20.0f * mf::vml_log10(srcv);
}

static float UNUSED
dB2lin(float srcv)
{
	typedef mathfuncs<float1> mf;
	const static float div20 = 0.05f;
	return mf::vml_exp10(srcv * div20);
}


#endif
	

union cvtv4df
{
	v4df val;
	v2df val2[2];
};

union cvtv4sf
{
	v4sf val;
	v2sf val2[2];
};

union cvtv8df
{
	v8df val;
	v4df val2[2];
	v2df val4[4];
};

union cvtv8sf
{
	v8sf val;
	v4sf val2[2];
	v2sf val4[4];
};




#endif

//
// vim: set ts=4:
//
